import getpass
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional

try:
    from vault_tools.client.jsonsecret import get_plaintext
    from vault_tools.paasta_secret import get_vault_client
    from vault_tools.gpg import TempGpgKeyring
    from vault_tools.paasta_secret import encrypt_secret
    from vault_tools.cert_tools import do_renew
    import hvac
except ImportError:

    def get_plaintext(*args: Any, **kwargs: Any) -> bytes:
        return b"No plain text available without vault_tools"

    def get_vault_client(*args: Any, **kwargs: Any) -> None:
        return None

    TempGpgKeyring = None

    def encrypt_secret(*args: Any, **kwargs: Any) -> None:
        return None

    def do_renew(*args: Any, **kwargs: Any) -> None:
        return None


from paasta_tools.secret_providers import BaseSecretProvider
from paasta_tools.utils import paasta_print
from paasta_tools.secret_tools import get_secret_name_from_ref


class SecretProvider(BaseSecretProvider):
    def __init__(
        self,
        soa_dir: Optional[str],
        service_name: Optional[str],
        cluster_names: List[str],
        vault_cluster_config: Dict[str, str] = {},
        vault_auth_method: str = "ldap",
        vault_token_file: str = "/root/.vault-token",
        vault_num_uses: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(soa_dir, service_name, cluster_names)
        self.vault_cluster_config = vault_cluster_config
        self.vault_auth_method = vault_auth_method
        self.vault_token_file = vault_token_file
        self.ecosystems = self.get_vault_ecosystems_for_clusters()
        self.clients: Mapping[str, hvac.Client] = {}
        if vault_auth_method == "ldap":
            username = getpass.getuser()
            password = getpass.getpass(
                "Please enter your LDAP password to auth with Vault\n"
            )
        else:
            username = None
            password = None
        for ecosystem in self.ecosystems:
            self.clients[ecosystem] = get_vault_client(
                ecosystem=ecosystem,
                num_uses=vault_num_uses,
                vault_auth_method=self.vault_auth_method,
                vault_token_file=self.vault_token_file,
                username=username,
                password=password,
            )

    def decrypt_environment(
        self, environment: Dict[str, str], **kwargs: Any
    ) -> Dict[str, str]:
        client = self.clients[self.ecosystems[0]]
        secret_environment = {}
        for k, v in environment.items():
            secret_name = get_secret_name_from_ref(v)
            secret_path = os.path.join(self.secret_dir, f"{secret_name}.json")
            secret = get_plaintext(
                client=client,
                env=self.ecosystems[0],
                path=secret_path,
                cache_enabled=False,
                cache_dir=None,
                cache_key=None,
                context=self.service_name,
            ).decode("utf-8")
            secret_environment[k] = secret
        return secret_environment

    def get_vault_ecosystems_for_clusters(self) -> List[str]:
        try:
            return list(
                {
                    self.vault_cluster_config[cluster_name]
                    for cluster_name in self.cluster_names
                }
            )
        except KeyError as e:
            paasta_print(
                "Cannot find a vault cluster for the %s paasta cluster. A mapping must exist "
                "in /etc/paasta so we contact the correct vault cluster to get/set secrets"
                % e
            )
            raise

    def write_secret(self, action: str, secret_name: str, plaintext: bytes) -> None:
        with TempGpgKeyring(overwrite=True):
            for ecosystem in self.ecosystems:
                client = self.clients[ecosystem]
                encrypt_secret(
                    client=client,
                    action=action,
                    ecosystem=ecosystem,
                    secret_name=secret_name,
                    soa_dir=self.soa_dir,
                    plaintext=plaintext,
                    service_name=self.service_name,
                    transit_key=self.encryption_key,
                )

    def decrypt_secret(self, secret_name: str) -> str:
        client = self.clients[self.ecosystems[0]]
        secret_path = os.path.join(self.secret_dir, f"{secret_name}.json")
        return get_plaintext(
            client=client,
            path=secret_path,
            env=self.ecosystems[0],
            cache_enabled=False,
            cache_key=None,
            cache_dir=None,
            context=self.service_name,
        ).decode("utf-8")

    def decrypt_secret_raw(self, secret_name: str) -> bytes:
        client = self.clients[self.ecosystems[0]]
        secret_path = os.path.join(self.secret_dir, f"{secret_name}.json")
        return get_plaintext(
            client=client,
            path=secret_path,
            env=self.ecosystems[0],
            cache_enabled=False,
            cache_key=None,
            cache_dir=None,
            context=self.service_name,
        )

    def get_secret_signature_from_data(self, data: Mapping[str, Any]) -> Optional[str]:
        ecosystem = self.ecosystems[0]
        if data["environments"][ecosystem]:
            return data["environments"][ecosystem]["signature"]
        else:
            return None

    def renew_issue_cert(self, pki_backend: str, ttl: str) -> None:
        client = self.clients[self.ecosystems[0]]
        user = getpass.getuser()
        pki_dir = os.path.expanduser("~/.paasta/pki")
        do_renew(
            client=client,
            pki_backend=pki_backend,
            role=user,
            cn=f"{user}.{self.ecosystems[0]}.paasta.yelp",
            cert_path=f"{pki_dir}/{self.ecosystems[0]}.crt",
            key_path=f"{pki_dir}/{self.ecosystems[0]}.key",
            ca_path=f"{pki_dir}/{self.ecosystems[0]}_ca.crt",
            cert_owner=user,
            cert_group="users",
            cert_mode="0600",
            ttl=ttl,
        )



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import sys
import time
from collections import namedtuple
from random import choice

from pysensu_yelp import Status

from paasta_tools import monitoring_tools
from paasta_tools.chronos_tools import compose_check_name_for_service_instance
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config

try:
    from scribereader import scribereader
except ImportError:
    scribereader = None


OOM_EVENTS_STREAM = "tmp_paasta_oom_events"

OOMEvent = namedtuple("OOMEvent", ["hostname", "container_id", "process_name"])


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=(
            "Check the %s stream and report to Sensu if"
            " there are any OOM events." % OOM_EVENTS_STREAM
        )
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-r",
        "--realert-every",
        dest="realert_every",
        type=int,
        default=1,
        help="Sensu 'realert_every' to use.",
    )
    parser.add_argument(
        "-s",
        "--superregion",
        dest="superregion",
        required=True,
        help="The superregion to read OOM events from.",
    )
    return parser.parse_args(args)


def read_oom_events_from_scribe(cluster, superregion, num_lines=1000):
    """Read the latest 'num_lines' lines from OOM_EVENTS_STREAM and iterate over them."""
    host_port = choice(scribereader.get_default_scribe_hosts(tail=True))
    stream = scribereader.get_stream_tailer(
        stream_name=OOM_EVENTS_STREAM,
        tailing_host=host_port["host"],
        tailing_port=host_port["port"],
        use_kafka=True,
        lines=num_lines,
        superregion=superregion,
    )
    for line in stream:
        try:
            j = json.loads(line)
            if j.get("cluster", "") == cluster:
                yield j
        except json.decoder.JSONDecodeError:
            pass


def latest_oom_events(cluster, superregion, interval=60):
    """
    :returns: {(service, instance): [OOMEvent, OOMEvent,...] }
              if the number of events > 0
    """
    start_timestamp = int(time.time()) - interval
    res = {}
    for e in read_oom_events_from_scribe(cluster, superregion):
        if e["timestamp"] > start_timestamp:
            key = (e["service"], e["instance"])
            res.setdefault(key, []).append(
                OOMEvent(
                    hostname=e.get("hostname", ""),
                    container_id=e.get("container_id", ""),
                    process_name=e.get("process_name", ""),
                )
            )
    return res


def compose_sensu_status(instance, oom_events, is_check_enabled):
    """
    :param instance: InstanceConfig
    :param oom_events: a list of OOMEvents
    :param is_check_enabled: boolean to indicate whether the check enabled for the instance
    """
    if not is_check_enabled:
        return (
            Status.OK,
            "This check is disabled for {}.{}.".format(
                instance.service, instance.instance
            ),
        )
    if len(oom_events) == 0:
        return (
            Status.OK,
            "No oom events for %s.%s in the last minute."
            % (instance.service, instance.instance),
        )
    else:
        return (
            Status.CRITICAL,
            "The Out Of Memory killer killed %d processes (%s) "
            "in the last minute in %s.%s containers."
            % (
                len(oom_events),
                ",".join(
                    sorted({e.process_name for e in oom_events if e.process_name})
                ),
                instance.service,
                instance.instance,
            ),
        )


def send_sensu_event(instance, oom_events, args):
    """
    :param instance: InstanceConfig
    :param oom_events: a list of OOMEvents
    """
    check_name = compose_check_name_for_service_instance(
        "oom-killer", instance.service, instance.instance
    )
    monitoring_overrides = instance.get_monitoring()
    status = compose_sensu_status(
        instance=instance,
        oom_events=oom_events,
        is_check_enabled=monitoring_overrides.get("check_oom_events", True),
    )
    monitoring_overrides.update(
        {
            "page": False,
            "ticket": False,
            "alert_after": "0m",
            "realert_every": args.realert_every,
            "runbook": "y/check-oom-events",
            "tip": "Try bumping the memory limit past %dMB" % instance.get_mem(),
        }
    )
    return monitoring_tools.send_event(
        service=instance.service,
        check_name=check_name,
        overrides=monitoring_overrides,
        status=status[0],
        output=status[1],
        soa_dir=instance.soa_dir,
    )


def main(sys_argv):
    args = parse_args(sys_argv[1:])
    cluster = load_system_paasta_config().get_cluster()
    victims = latest_oom_events(cluster, args.superregion)
    for (service, instance) in get_services_for_cluster(cluster, soa_dir=args.soa_dir):
        try:
            instance_config = get_instance_config(
                service=service,
                instance=instance,
                cluster=cluster,
                load_deployments=False,
                soa_dir=args.soa_dir,
            )
            oom_events = victims.get((service, instance), [])
            send_sensu_event(instance_config, oom_events, args)
        except NotImplementedError:  # When instance_type is not supported by get_instance_config
            pass


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import shlex
import subprocess
import sys

from paasta_tools.cli.utils import get_container_name
from paasta_tools.cli.utils import get_subparser
from paasta_tools.cli.utils import get_task_from_instance
from paasta_tools.cli.utils import PaastaTaskNotFound


def add_subparser(subparsers):
    new_parser = get_subparser(
        description="'paasta docker_exec' works by picking a container running your service "
        "at random. It then runs docker exec -ti <container_id> <commands> "
        "where commands are those that you specify",
        help_text="Docker exec against a container running your service",
        command="docker_exec",
        function=paasta_docker_exec,
        subparsers=subparsers,
    )
    new_parser.add_argument(
        "--user", help="User to exec into the container as (e.g. root)"
    )
    new_parser.add_argument(
        "exec_command",
        help="Command to append to docker docker_exec",
        nargs="?",
        default="/bin/bash",
    )


def paasta_docker_exec(args):
    try:
        task = get_task_from_instance(
            cluster=args.cluster,
            service=args.service,
            instance=args.instance,
            slave_hostname=args.host,
            task_id=args.mesos_id,
        )
    except PaastaTaskNotFound:
        sys.exit(1)
    container = get_container_name(task)
    slave = task.slave["hostname"]

    user_args = ("--user", args.user) if args.user else ()
    command = (
        "sudo",
        "docker",
        "exec",
        "-ti",
        *user_args,
        container,
        *shlex.split(args.exec_command),
    )
    command = " ".join(shlex.quote(arg) for arg in command)
    subprocess.call(["ssh", "-o", "LogLevel=QUIET", "-tA", slave, command])



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.cli.utils import execute_paasta_serviceinit_on_remote_master
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_instances
from paasta_tools.utils import _log_audit
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def add_subparser(subparsers):
    status_parser = subparsers.add_parser(
        "emergency-restart",
        help="Restarts a PaaSTA service instance in an emergency",
        description=(
            "'paasta emergency-restart' is useful in situations where the operator "
            "needs to bypass the normal git-based control plan, and needs to interact "
            "with the underlying APIs directly. For example, in an emergency situation "
            "it may be necessary to restart a Marathon service without doing a 'full bounce'."
            "'emergency-restart' can do this, but at the cost of the safety of the normal "
            "bouncing procedures. In other words, and emergency-restart is fast, but not safe "
            "and will cause dropped traffic.\n\n"
        ),
    )
    status_parser.add_argument(
        "-s",
        "--service",
        help="Service that you want to restart. Like 'example_service'.",
    ).completer = lazy_choices_completer(list_services)
    status_parser.add_argument(
        "-i",
        "--instance",
        help="Instance of the service that you want to restart. Like 'main' or 'canary'.",
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    status_parser.add_argument(
        "-c",
        "--cluster",
        help="The PaaSTA cluster that has the service you want to restart. Like 'norcal-prod'.",
        required=True,
    ).completer = lazy_choices_completer(list_clusters)
    status_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    status_parser.set_defaults(command=paasta_emergency_restart)


def paasta_emergency_restart(args):
    """Performs an emergency restart on a given service instance on a given cluster

    Warning: This command is only intended to be used in an emergency.
    It should not be needed in normal circumstances.
    """
    service = figure_out_service_name(args, args.soa_dir)
    system_paasta_config = load_system_paasta_config()
    paasta_print(
        "Performing an emergency restart on %s...\n"
        % compose_job_id(service, args.instance)
    )
    return_code, output = execute_paasta_serviceinit_on_remote_master(
        subcommand="restart",
        cluster=args.cluster,
        service=args.service,
        instances=args.instance,
        system_paasta_config=system_paasta_config,
    )
    _log_audit(
        action="emergency-restart",
        service=args.service,
        cluster=args.cluster,
        instance=args.instance,
    )

    paasta_print("Output: %s" % output)
    paasta_print("%s" % "\n".join(paasta_emergency_restart.__doc__.splitlines()[-7:]))
    paasta_print("Run this to see the status:")
    paasta_print(f"paasta status --service {service} --clusters {args.cluster}")

    return return_code



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import itertools
import math
from collections import Counter
from collections import namedtuple
from collections import OrderedDict
from typing import Any
from typing import Callable
from typing import Mapping
from typing import NamedTuple
from typing import Sequence
from typing import Tuple
from typing import TypeVar

import a_sync
from humanize import naturalsize
from kubernetes.client import V1Node
from mypy_extensions import TypedDict
from typing_extensions import Counter as _Counter

from paasta_tools import chronos_tools
from paasta_tools.kubernetes_tools import get_all_nodes
from paasta_tools.kubernetes_tools import get_all_pods
from paasta_tools.kubernetes_tools import get_pod_status
from paasta_tools.kubernetes_tools import is_node_ready
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import list_all_deployments
from paasta_tools.kubernetes_tools import maybe_add_yelp_prefix
from paasta_tools.kubernetes_tools import PodStatus
from paasta_tools.marathon_tools import get_all_marathon_apps
from paasta_tools.marathon_tools import MarathonClient
from paasta_tools.mesos.master import MesosMetrics
from paasta_tools.mesos.master import MesosState
from paasta_tools.mesos_maintenance import MAINTENANCE_ROLE
from paasta_tools.mesos_tools import get_all_tasks_from_state
from paasta_tools.mesos_tools import get_mesos_quorum
from paasta_tools.mesos_tools import get_number_of_mesos_masters
from paasta_tools.mesos_tools import get_zookeeper_host_path
from paasta_tools.mesos_tools import is_task_terminal
from paasta_tools.mesos_tools import MesosResources
from paasta_tools.mesos_tools import MesosTask
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import print_with_indent


class ResourceInfo(namedtuple("ResourceInfo", ["cpus", "mem", "disk", "gpus"])):
    def __new__(cls, cpus, mem, disk, gpus=0):
        return super().__new__(cls, cpus, mem, disk, gpus)


class HealthCheckResult(NamedTuple):
    message: str
    healthy: bool


class ResourceUtilization(NamedTuple):
    metric: str
    total: int
    free: int


HIGH_QUEUE_GAUGE = "org.apache.mesos.chronos.scheduler.jobs.TaskManager.highQueueSize"
QUEUE_GAUGE = "org.apache.mesos.chronos.scheduler.jobs.TaskManager.queueSize"


def get_num_masters() -> int:
    """ Gets the number of masters from mesos state """
    zookeeper_host_path = get_zookeeper_host_path()
    return get_number_of_mesos_masters(
        zookeeper_host_path.host, zookeeper_host_path.path
    )


def get_mesos_cpu_status(
    metrics: MesosMetrics, mesos_state: MesosState
) -> Tuple[int, int, int]:
    """Takes in the mesos metrics and analyzes them, returning the status.

    :param metrics: mesos metrics dictionary.
    :param mesos_state: mesos state dictionary.
    :returns: Tuple of total, used, and available CPUs.
    """

    total = metrics["master/cpus_total"]
    used = metrics["master/cpus_used"]

    for slave in mesos_state["slaves"]:
        used += reserved_maintenence_resources(slave["reserved_resources"])["cpus"]

    available = total - used
    return total, used, available


def get_kube_cpu_status(nodes: Sequence[V1Node],) -> Tuple[int, int, int]:
    """Takes in the list of Kubernetes nodes and analyzes them, returning the status.

    :param nodes: list of Kubernetes nodes.
    :returns: Tuple of total, used, and available CPUs.
    """

    total = 0
    available = 0
    for node in nodes:
        available += suffixed_number_value(node.status.allocatable["cpu"])
        total += suffixed_number_value(node.status.capacity["cpu"])

    used = total - available
    return total, used, available


def get_mesos_memory_status(
    metrics: MesosMetrics, mesos_state: MesosState
) -> Tuple[int, int, int]:
    """Takes in the mesos metrics and analyzes them, returning the status.

    :param metrics: mesos metrics dictionary.
    :param mesos_state: mesos state dictionary.
    :returns: Tuple of total, used, and available memory in Mi.
    """
    total = metrics["master/mem_total"]
    used = metrics["master/mem_used"]

    for slave in mesos_state["slaves"]:
        used += reserved_maintenence_resources(slave["reserved_resources"])["mem"]

    available = total - used

    return total, used, available


def get_kube_memory_status(nodes: Sequence[V1Node],) -> Tuple[int, int, int]:
    """Takes in the list of Kubernetes nodes and analyzes them, returning the status.

    :param nodes: list of Kubernetes nodes.
    :returns: Tuple of total, used, and available memory in Mi.
    """
    total = 0
    available = 0
    for node in nodes:
        available += suffixed_number_value(node.status.allocatable["memory"])
        total += suffixed_number_value(node.status.capacity["memory"])

    total //= 1024 * 1024
    available //= 1024 * 1024
    used = total - available
    return total, used, available


def get_mesos_disk_status(
    metrics: MesosMetrics, mesos_state: MesosState
) -> Tuple[int, int, int]:
    """Takes in the mesos metrics and analyzes them, returning the status.

    :param metrics: mesos metrics dictionary.
    :param mesos_state: mesos state dictionary.
    :returns: Tuple of total, used, and available disk space in Mi.
    """

    total = metrics["master/disk_total"]
    used = metrics["master/disk_used"]

    for slave in mesos_state["slaves"]:
        used += reserved_maintenence_resources(slave["reserved_resources"])["disk"]

    available = total - used
    return total, used, available


def get_kube_disk_status(nodes: Sequence[V1Node],) -> Tuple[int, int, int]:
    """Takes in the list of Kubernetes nodes and analyzes them, returning the status.

    :param nodes: list of Kubernetes nodes.
    :returns: Tuple of total, used, and available disk space in Mi.
    """

    total = 0
    available = 0
    for node in nodes:
        available += suffixed_number_value(node.status.allocatable["ephemeral-storage"])
        total += suffixed_number_value(node.status.capacity["ephemeral-storage"])

    total //= 1024 * 1024
    available //= 1024 * 1024
    used = total - available
    return total, used, available


def get_mesos_gpu_status(
    metrics: MesosMetrics, mesos_state: MesosState
) -> Tuple[int, int, int]:
    """Takes in the mesos metrics and analyzes them, returning gpus status.

    :param metrics: mesos metrics dictionary.
    :param mesos_state: mesos state dictionary.
    :returns: Tuple of total, used, and available GPUs.
    """
    total = metrics["master/gpus_total"]
    used = metrics["master/gpus_used"]

    for slave in mesos_state["slaves"]:
        used += reserved_maintenence_resources(slave["reserved_resources"])["gpus"]

    available = total - used
    return total, used, available


def get_kube_gpu_status(nodes: Sequence[V1Node],) -> Tuple[int, int, int]:
    """Takes in the list of Kubernetes nodes and analyzes them, returning the status.

    :param nodes: list of Kubernetes nodes.
    :returns: Tuple of total, used, and available GPUs.
    """

    total = 0
    available = 0
    for node in nodes:
        available += suffixed_number_value(
            node.status.allocatable.get("nvidia.com/gpu", "0")
        )
        total += suffixed_number_value(node.status.capacity.get("nvidia.com/gpu", "0"))

    used = total - available
    return total, used, available


def filter_mesos_state_metrics(dictionary: Mapping[str, Any]) -> Mapping[str, Any]:
    valid_keys = ["cpus", "mem", "disk", "gpus"]
    return {key: value for (key, value) in dictionary.items() if key in valid_keys}


def filter_kube_resources(dictionary: Mapping[str, str]) -> Mapping[str, str]:
    valid_keys = ["cpu", "memory", "ephemeral-storage", "nvidia.com/gpu"]
    return {key: value for (key, value) in dictionary.items() if key in valid_keys}


def healthcheck_result_for_resource_utilization(
    resource_utilization: ResourceUtilization, threshold: int
) -> HealthCheckResult:
    """ Given a resource data dict, assert that cpu
    data is ok.

    :param resource_utilization: the resource_utilization tuple to check
    :returns: a HealthCheckResult
    """
    try:
        utilization = percent_used(
            resource_utilization.total,
            resource_utilization.total - resource_utilization.free,
        )
    except ZeroDivisionError:
        utilization = 0
    message = "{}: {:.2f}/{:.2f}({:.2f}%) used. Threshold ({:.2f}%)".format(
        resource_utilization.metric,
        float(resource_utilization.total - resource_utilization.free),
        resource_utilization.total,
        utilization,
        threshold,
    )
    healthy = utilization <= threshold
    return HealthCheckResult(message=message, healthy=healthy)


def quorum_ok(masters: int, quorum: int) -> bool:
    return masters >= quorum


def check_threshold(percent_used: float, threshold: int) -> bool:
    return (100 - percent_used) > threshold


def percent_used(total: float, used: float) -> float:
    return round(used / float(total) * 100.0, 2)


def assert_cpu_health(
    cpu_status: Tuple[int, int, int], threshold: int = 10
) -> HealthCheckResult:
    total, used, available = cpu_status
    try:
        perc_used = percent_used(total, used)
    except ZeroDivisionError:
        return HealthCheckResult(
            message="Error reading total available cpu from mesos!", healthy=False
        )

    if check_threshold(perc_used, threshold):
        return HealthCheckResult(
            message="CPUs: %.2f / %d in use (%s)"
            % (used, total, PaastaColors.green("%.2f%%" % perc_used)),
            healthy=True,
        )
    else:
        return HealthCheckResult(
            message="CRITICAL: Less than %d%% CPUs available. (Currently using %.2f%% of %d)"
            % (threshold, perc_used, total),
            healthy=False,
        )


def assert_memory_health(
    memory_status: Tuple[int, int, int], threshold: int = 10
) -> HealthCheckResult:
    total: float
    used: float
    total, used, _ = memory_status

    total /= 1024
    used /= 1024

    try:
        perc_used = percent_used(total, used)
    except ZeroDivisionError:
        return HealthCheckResult(
            message="Error reading total available memory from mesos!", healthy=False
        )

    if check_threshold(perc_used, threshold):
        return HealthCheckResult(
            message="Memory: %0.2f / %0.2fGB in use (%s)"
            % (used, total, PaastaColors.green("%.2f%%" % perc_used)),
            healthy=True,
        )
    else:
        return HealthCheckResult(
            message="CRITICAL: Less than %d%% memory available. (Currently using %.2f%% of %.2fGB)"
            % (threshold, perc_used, total),
            healthy=False,
        )


def assert_disk_health(
    disk_status: Tuple[int, int, int], threshold: int = 10
) -> HealthCheckResult:
    total: float
    used: float
    total, used, _ = disk_status

    total /= 1024
    used /= 1024

    try:
        perc_used = percent_used(total, used)
    except ZeroDivisionError:
        return HealthCheckResult(
            message="Error reading total available disk from mesos!", healthy=False
        )

    if check_threshold(perc_used, threshold):
        return HealthCheckResult(
            message="Disk: %0.2f / %0.2fGB in use (%s)"
            % (used, total, PaastaColors.green("%.2f%%" % perc_used)),
            healthy=True,
        )
    else:
        return HealthCheckResult(
            message="CRITICAL: Less than %d%% disk available. (Currently using %.2f%%)"
            % (threshold, perc_used),
            healthy=False,
        )


def assert_gpu_health(
    gpu_status: Tuple[int, int, int], threshold: int = 0
) -> HealthCheckResult:
    total, used, available = gpu_status

    if math.isclose(total, 0):
        # assume that no gpus is healthy since most machines don't have them
        return HealthCheckResult(message="No GPUs found!", healthy=True)
    else:
        perc_used = percent_used(total, used)

    if check_threshold(perc_used, threshold):
        # only whole gpus can be used
        return HealthCheckResult(
            message="GPUs: %d / %d in use (%s)"
            % (used, total, PaastaColors.green("%.2f%%" % perc_used)),
            healthy=True,
        )
    else:
        return HealthCheckResult(
            message="CRITICAL: Less than %d%% GPUs available. (Currently using %.2f%% of %d)"
            % (threshold, perc_used, total),
            healthy=False,
        )


def assert_mesos_tasks_running(metrics: MesosMetrics,) -> HealthCheckResult:
    running = metrics["master/tasks_running"]
    staging = metrics["master/tasks_staging"]
    starting = metrics["master/tasks_starting"]
    return HealthCheckResult(
        message="Tasks: running: %d staging: %d starting: %d"
        % (running, staging, starting),
        healthy=True,
    )


def assert_kube_pods_running(kube_client: KubeClient,) -> HealthCheckResult:
    statuses = [get_pod_status(pod) for pod in get_all_pods(kube_client)]
    running = statuses.count(PodStatus.RUNNING)
    pending = statuses.count(PodStatus.PENDING)
    failed = statuses.count(PodStatus.FAILED)
    healthy = running > 0
    return HealthCheckResult(
        message=f"Pods: running: {running} pending: {pending} failed: {failed}",
        healthy=healthy,
    )


def assert_no_duplicate_frameworks(
    state: MesosState, framework_list: Sequence[str] = ["marathon", "chronos"]
) -> HealthCheckResult:
    """A function which asserts that there are no duplicate frameworks running, where
    frameworks are identified by their name.

    Note the extra spaces in the output strings: this is to account for the extra indentation
    we add, so we can have:

        frameworks:
          framework: marathon count: 1

    :param state: the state info from the Mesos master
    :returns: a tuple containing (output, ok): output is a log of the state of frameworks, ok a boolean
        indicating if there are any duplicate frameworks.
    """
    output = ["Frameworks:"]
    status = True
    frameworks = state["frameworks"]
    for name in framework_list:
        shards = [x["name"] for x in frameworks if x["name"].startswith(name)]
        for framework, count in OrderedDict(sorted(Counter(shards).items())).items():
            if count > 1:
                status = False
                output.append(
                    "    CRITICAL: There are %d connected %s frameworks! "
                    "(Expected 1)" % (count, framework)
                )
        output.append("    Framework: %s count: %d" % (name, len(shards)))

    return HealthCheckResult(message=("\n").join(output), healthy=status)


def assert_frameworks_exist(
    state: MesosState, expected: Sequence[str]
) -> HealthCheckResult:
    frameworks = [f["name"] for f in state["frameworks"]]
    not_found = []
    ok = True

    for f in expected:
        if f not in frameworks:
            ok = False
            not_found.append(f)

    if ok:
        return HealthCheckResult(message="all expected frameworks found", healthy=ok)
    else:
        return HealthCheckResult(
            message="CRITICAL: framework(s) %s not found" % ", ".join(not_found),
            healthy=ok,
        )


def get_mesos_slaves_health_status(metrics: MesosMetrics,) -> Tuple[int, int]:
    return metrics["master/slaves_active"], metrics["master/slaves_inactive"]


def get_kube_nodes_health_status(nodes: Sequence[V1Node],) -> Tuple[int, int]:
    statuses = [is_node_ready(node) for node in nodes]
    return statuses.count(True), statuses.count(False)


def assert_nodes_health(nodes_health_status: Tuple[int, int],) -> HealthCheckResult:
    active, inactive = nodes_health_status
    healthy = active > 0
    return HealthCheckResult(
        message="Nodes: active: %d inactive: %d" % (active, inactive), healthy=healthy
    )


def assert_quorum_size() -> HealthCheckResult:
    masters, quorum = get_num_masters(), a_sync.block(get_mesos_quorum)
    if quorum_ok(masters, quorum):
        return HealthCheckResult(
            message="Quorum: masters: %d configured quorum: %d " % (masters, quorum),
            healthy=True,
        )
    else:
        return HealthCheckResult(
            message="CRITICAL: Number of masters (%d) less than configured quorum(%d)."
            % (masters, quorum),
            healthy=False,
        )


_KeyFuncRetT = Sequence[Tuple[str, str]]


class _SlaveT(TypedDict):
    id: str
    resources: MesosResources
    reserved_resources: MesosResources
    attributes: Mapping[str, str]


_GenericNodeT = TypeVar("_GenericNodeT", _SlaveT, V1Node)

_GenericNodeGroupingFunctionT = Callable[[_GenericNodeT], _KeyFuncRetT]

_GenericNodeFilterFunctionT = Callable[[_GenericNodeT], bool]

_GenericNodeSortFunctionT = Callable[[Sequence[_GenericNodeT]], Sequence[_GenericNodeT]]


def key_func_for_attribute(attribute: str,) -> Callable[[_SlaveT], str]:
    """ Return a closure that given a slave, will return the value of a specific
    attribute.

    :param attribute: the attribute to inspect in the slave
    :returns: a closure, which takes a slave and returns the value of an attribute
    """

    def key_func(slave):
        return slave["attributes"].get(attribute, "unknown")

    return key_func


def key_func_for_attribute_multi(
    attributes: Sequence[str],
) -> _GenericNodeGroupingFunctionT:
    """ Return a closure that given a slave, will return the value of a list of
    attributes, compiled into a hashable tuple

    :param attributes: the attributes to inspect in the slave
    :returns: a closure, which takes a slave and returns the value of those attributes
    """

    def get_attribute(slave, attribute):
        if attribute == "hostname":
            return slave["hostname"]
        else:
            return slave["attributes"].get(attribute, "unknown")

    def key_func(slave):
        return tuple((a, get_attribute(slave, a)) for a in attributes)

    return key_func


def key_func_for_attribute_multi_kube(
    attributes: Sequence[str],
) -> Callable[[V1Node], _KeyFuncRetT]:
    """ Return a closure that given a node, will return the value of a list of
    attributes, compiled into a hashable tuple

    :param attributes: the attributes to inspect in the slave
    :returns: a closure, which takes a node and returns the value of those attributes
    """

    def get_attribute(node, attribute):
        return node.metadata.labels.get(maybe_add_yelp_prefix(attribute), "unknown")

    def key_func(node):
        return tuple((a, get_attribute(node, a)) for a in attributes)

    return key_func


def sort_func_for_attributes(attributes: Sequence[str],) -> _GenericNodeSortFunctionT:
    def sort(slaves):
        for attribute in attributes:
            slaves = sorted(slaves, key=key_func_for_attribute(attribute))
        return slaves

    return sort


def group_slaves_by_key_func(
    key_func: _GenericNodeGroupingFunctionT,
    slaves: Sequence[_GenericNodeT],
    sort_func: _GenericNodeSortFunctionT = None,
) -> Mapping[_KeyFuncRetT, Sequence[_GenericNodeT]]:
    """ Given a function for grouping slaves, return a
    dict where keys are the unique values returned by
    the key_func and the values are all those slaves which
    have that specific value.

    :param key_func: a function which consumes a slave and returns a value
    :param slaves: a list of slaves
    :returns: a dict of key: [slaves]
    """
    sorted_slaves: Sequence[_GenericNodeT]
    if sort_func is None:
        sorted_slaves = sorted(slaves, key=key_func)
    else:
        sorted_slaves = sort_func(slaves)

    return {k: list(v) for k, v in itertools.groupby(sorted_slaves, key=key_func)}


class ResourceUtilizationDict(TypedDict):
    free: ResourceInfo
    total: ResourceInfo
    slave_count: int


def calculate_resource_utilization_for_slaves(
    slaves: Sequence[_SlaveT], tasks: Sequence[MesosTask]
) -> ResourceUtilizationDict:
    """ Given a list of slaves and a list of tasks, calculate the total available
    resource available in that list of slaves, and the resources consumed by tasks
    running on those slaves.

    :param slaves: a list of slaves to calculate resource usage for
    :param tasks: the list of tasks running in the mesos cluster
    :returns: a dict, containing keys for "free" and "total" resources. Each of these keys
    is a ResourceInfo tuple, exposing a number for cpu, disk and mem.
    """
    resource_total_dict: _Counter[str] = Counter()
    for slave in slaves:
        filtered_resources = filter_mesos_state_metrics(slave["resources"])
        resource_total_dict.update(Counter(filtered_resources))
    resource_free_dict = copy.deepcopy(resource_total_dict)
    for task in tasks:
        task_resources = task["resources"]
        resource_free_dict.subtract(Counter(filter_mesos_state_metrics(task_resources)))
    for slave in slaves:
        filtered_resources = filter_mesos_state_metrics(
            reserved_maintenence_resources(slave["reserved_resources"])
        )
        resource_free_dict.subtract(Counter(filtered_resources))
    return {
        "free": ResourceInfo(
            cpus=resource_free_dict["cpus"],
            disk=resource_free_dict["disk"],
            mem=resource_free_dict["mem"],
            gpus=resource_free_dict.get("gpus", 0),
        ),
        "total": ResourceInfo(
            cpus=resource_total_dict["cpus"],
            disk=resource_total_dict["disk"],
            mem=resource_total_dict["mem"],
            gpus=resource_total_dict.get("gpus", 0),
        ),
        "slave_count": len(slaves),
    }


_IEC_NUMBER_SUFFIXES = {
    "Ki": 1024,
    "Mi": 1024 ** 2,
    "Gi": 1024 ** 3,
    "Ti": 1024 ** 4,
    "Pi": 1024 ** 5,
}


def suffixed_number_value(s: str) -> int:
    suff = s[-2:]
    if suff in _IEC_NUMBER_SUFFIXES:
        return int(s[:-2]) * _IEC_NUMBER_SUFFIXES[suff]
    else:
        return int(s)


def suffixed_number_dict_values(d: Mapping[Any, str]) -> Mapping[Any, int]:
    return {k: suffixed_number_value(v) for k, v in d.items()}


def calculate_resource_utilization_for_kube_nodes(
    nodes: Sequence[V1Node],
) -> ResourceUtilizationDict:
    """ Given a list of Kubernetes nodes, calculate the total available
    resource available and the resources consumed in that list of nodes.

    :param nodes: a list of Kubernetes nodes to calculate resource usage for
    :returns: a dict, containing keys for "free" and "total" resources. Each of these keys
    is a ResourceInfo tuple, exposing a number for cpu, disk and mem.
    """
    resource_total_dict: _Counter[str] = Counter()
    for node in nodes:
        filtered_resources = filter_kube_resources(node.status.capacity)
        resource_total_dict.update(
            Counter(suffixed_number_dict_values(filtered_resources))
        )
    resource_free_dict: _Counter[str] = Counter()
    for node in nodes:
        filtered_resources = filter_kube_resources(node.status.allocatable)
        resource_free_dict.update(
            Counter(suffixed_number_dict_values(filtered_resources))
        )
    return {
        "free": ResourceInfo(
            cpus=resource_free_dict["cpu"],
            disk=resource_free_dict["ephemeral-storage"] / (1024 ** 2),
            mem=resource_free_dict["memory"] / (1024 ** 2),
            gpus=resource_free_dict.get("nvidia.com/gpu", 0),
        ),
        "total": ResourceInfo(
            cpus=resource_total_dict["cpu"],
            disk=resource_total_dict["ephemeral-storage"] / (1024 ** 2),
            mem=resource_total_dict["memory"] / (1024 ** 2),
            gpus=resource_total_dict.get("nvidia.com/gpu", 0),
        ),
        "slave_count": len(nodes),
    }


def filter_tasks_for_slaves(
    slaves: Sequence[_SlaveT], tasks: Sequence[MesosTask]
) -> Sequence[MesosTask]:
    """ Given a list of slaves and a list of tasks, return a filtered
    list of tasks, where those returned belong to slaves in the list of
    slaves

    :param slaves: the list of slaves which the tasks provided should be
    running on.
    :param tasks: the tasks to filter :returns: a list of tasks,
    identical to that provided by the tasks param, but with only those where
    the task is running on one of the provided slaves included.
    """
    slave_ids = [slave["id"] for slave in slaves]
    return [task for task in tasks if task["slave_id"] in slave_ids]


def make_filter_slave_func(
    attribute: str, values: Sequence[str]
) -> _GenericNodeFilterFunctionT:
    def filter_func(slave):
        return slave["attributes"].get(attribute, None) in values

    return filter_func


def filter_slaves(
    slaves: Sequence[_GenericNodeT], filters: Sequence[_GenericNodeFilterFunctionT]
) -> Sequence[_GenericNodeT]:
    """ Filter slaves by attributes

    :param slaves: list of slaves to filter
    :param filters: list of functions that take a slave and return whether the
    slave should be included
    :returns: list of slaves that return true for all the filters
    """
    if filters is None:
        return slaves
    return [s for s in slaves if all([f(s) for f in filters])]


def get_resource_utilization_by_grouping(
    grouping_func: _GenericNodeGroupingFunctionT,
    mesos_state: MesosState,
    filters: Sequence[_GenericNodeFilterFunctionT] = [],
    sort_func: _GenericNodeSortFunctionT = None,
) -> Mapping[_KeyFuncRetT, ResourceUtilizationDict]:
    """ Given a function used to group slaves and mesos state, calculate
    resource utilization for each value of a given attribute.

    :grouping_func: a function that given a slave, will return the value of an
    attribute to group by.
    :param mesos_state: the mesos state
    :param filters: filters to apply to the slaves in the calculation, with
    filtering preformed by filter_slaves
    :param sort_func: a function that given a list of slaves, will return the
    sorted list of slaves.
    :returns: a dict of {attribute_value: resource_usage}, where resource usage
    is the dict returned by ``calculate_resource_utilization_for_slaves`` for
    slaves grouped by attribute value.
    """
    slaves: Sequence[_SlaveT] = mesos_state.get("slaves", [])
    slaves = filter_slaves(slaves, filters)
    if not has_registered_slaves(mesos_state):
        raise ValueError("There are no slaves registered in the mesos state.")

    tasks = get_all_tasks_from_state(mesos_state, include_orphans=True)
    non_terminal_tasks = [task for task in tasks if not is_task_terminal(task)]
    slave_groupings = group_slaves_by_key_func(grouping_func, slaves, sort_func)

    return {
        attribute_value: calculate_resource_utilization_for_slaves(
            slaves=slaves, tasks=filter_tasks_for_slaves(slaves, non_terminal_tasks)
        )
        for attribute_value, slaves in slave_groupings.items()
    }


def get_resource_utilization_by_grouping_kube(
    grouping_func: _GenericNodeGroupingFunctionT,
    kube_client: KubeClient,
    filters: Sequence[_GenericNodeFilterFunctionT] = [],
    sort_func: _GenericNodeSortFunctionT = None,
) -> Mapping[_KeyFuncRetT, ResourceUtilizationDict]:
    """ Given a function used to group nodes, calculate resource utilization
    for each value of a given attribute.

    :grouping_func: a function that given a node, will return the value of an
    attribute to group by.
    :param kube_client: the Kubernetes client
    :param filters: filters to apply to the nodes in the calculation, with
    filtering preformed by filter_slaves
    :param sort_func: a function that given a list of nodes, will return the
    sorted list of nodes.
    :returns: a dict of {attribute_value: resource_usage}, where resource usage
    is the dict returned by ``calculate_resource_utilization_for_kube_nodes`` for
    nodes grouped by attribute value.
    """
    nodes: Sequence[V1Node] = get_all_nodes(kube_client)
    nodes = filter_slaves(nodes, filters)
    if len(nodes) == 0:
        raise ValueError("There are no nodes registered in the Kubernetes.")

    node_groupings = group_slaves_by_key_func(grouping_func, nodes, sort_func)

    return {
        attribute_value: calculate_resource_utilization_for_kube_nodes(nodes)
        for attribute_value, nodes in node_groupings.items()
    }


def resource_utillizations_from_resource_info(
    total: ResourceInfo, free: ResourceInfo
) -> Sequence[ResourceUtilization]:
    """
    Given two ResourceInfo tuples, one for total and one for free,
    create a ResourceUtilization tuple for each metric in the ResourceInfo.
    :param total:
    :param free:
    :returns: ResourceInfo for a metric
    """
    return [
        ResourceUtilization(metric=field, total=total[index], free=free[index])
        for index, field in enumerate(ResourceInfo._fields)
    ]


def has_registered_slaves(mesos_state: MesosState,) -> bool:
    """ Return a boolean indicating if there are any slaves registered
    to the master according to the mesos state.
    :param mesos_state: the mesos state from the master
    :returns: a boolean, indicating if there are > 0 slaves
    """
    return len(mesos_state.get("slaves", [])) > 0


def get_mesos_resource_utilization_health(
    mesos_metrics: MesosMetrics, mesos_state: MesosState
) -> Sequence[HealthCheckResult]:
    """Perform healthchecks against mesos metrics.
    :param mesos_metrics: a dict exposing the mesos metrics described in
    https://mesos.apache.org/documentation/latest/monitoring/
    :returns: a list of HealthCheckResult tuples
    """
    return [
        assert_cpu_health(get_mesos_cpu_status(mesos_metrics, mesos_state)),
        assert_memory_health(get_mesos_memory_status(mesos_metrics, mesos_state)),
        assert_disk_health(get_mesos_disk_status(mesos_metrics, mesos_state)),
        assert_gpu_health(get_mesos_gpu_status(mesos_metrics, mesos_state)),
        assert_mesos_tasks_running(mesos_metrics),
        assert_nodes_health(get_mesos_slaves_health_status(mesos_metrics)),
    ]


def get_kube_resource_utilization_health(
    kube_client: KubeClient,
) -> Sequence[HealthCheckResult]:
    """Perform healthchecks against Kubernetes.
    :param kube_client: the KUbernetes client
    :returns: a list of HealthCheckResult tuples
    """

    nodes = get_all_nodes(kube_client)

    return [
        assert_cpu_health(get_kube_cpu_status(nodes)),
        assert_memory_health(get_kube_memory_status(nodes)),
        assert_disk_health(get_kube_disk_status(nodes)),
        assert_gpu_health(get_kube_gpu_status(nodes)),
        assert_nodes_health(get_kube_nodes_health_status(nodes)),
    ]


def get_mesos_state_status(mesos_state: MesosState,) -> Sequence[HealthCheckResult]:
    """Perform healthchecks against mesos state.
    :param mesos_state: a dict exposing the mesos state described in
    https://mesos.apache.org/documentation/latest/endpoints/master/state.json/
    :returns: a list of HealthCheckResult tuples
    """
    return [
        assert_quorum_size(),
        assert_no_duplicate_frameworks(
            state=mesos_state, framework_list=["marathon", "chronos"]
        ),
    ]


def run_healthchecks_with_param(
    param: Any,
    healthcheck_functions: Sequence[Callable[..., HealthCheckResult]],
    format_options: Mapping[str, Any] = {},
) -> Sequence[HealthCheckResult]:
    return [
        healthcheck(param, **format_options) for healthcheck in healthcheck_functions
    ]


def assert_marathon_apps(clients: Sequence[MarathonClient],) -> HealthCheckResult:
    num_apps = [len(get_all_marathon_apps(c)) for c in clients]
    if sum(num_apps) < 1:
        return HealthCheckResult(
            message="CRITICAL: No marathon apps running", healthy=False
        )
    else:
        return HealthCheckResult(
            message="marathon apps: %10d" % sum(num_apps), healthy=True
        )


def assert_marathon_tasks(clients: Sequence[MarathonClient],) -> HealthCheckResult:
    num_tasks = [len(c.list_tasks()) for c in clients]
    return HealthCheckResult(
        message="marathon tasks: %9d" % sum(num_tasks), healthy=True
    )


def assert_marathon_deployments(
    clients: Sequence[MarathonClient],
) -> HealthCheckResult:
    num_deployments = [len(c.list_deployments()) for c in clients]
    return HealthCheckResult(
        message="marathon deployments: %3d" % sum(num_deployments), healthy=True
    )


def assert_kube_deployments(kube_client: KubeClient,) -> HealthCheckResult:
    num_deployments = len(list_all_deployments(kube_client))
    return HealthCheckResult(
        message=f"Kubernetes deployments: {num_deployments:>3}", healthy=True
    )


def get_marathon_status(
    clients: Sequence[MarathonClient],
) -> Sequence[HealthCheckResult]:
    """ Gathers information about marathon.
    :return: string containing the status.  """
    return run_healthchecks_with_param(
        clients,
        [assert_marathon_apps, assert_marathon_tasks, assert_marathon_deployments],
    )


def assert_chronos_scheduled_jobs(client):
    """
    :returns: a tuple of a string and a bool containing representing if it is ok or not
    """
    num_jobs = len(chronos_tools.filter_enabled_jobs(client.list()))
    healthy = num_jobs != 0
    return HealthCheckResult(
        message="Enabled chronos jobs: %d" % num_jobs, healthy=healthy
    )


def assert_chronos_queued_jobs(client):
    high_priority_queue_size = client.metrics()["gauges"][HIGH_QUEUE_GAUGE]["value"]
    normal_priority_queue_size = client.metrics()["gauges"][QUEUE_GAUGE]["value"]
    all_jobs_queued = high_priority_queue_size + normal_priority_queue_size
    num_jobs = len(chronos_tools.filter_enabled_jobs(client.list()))

    try:
        perc_used = percent_used(num_jobs, all_jobs_queued)
    except ZeroDivisionError:
        perc_used = 0
    return HealthCheckResult(
        message=f"Jobs Queued: {all_jobs_queued} ({perc_used}%)", healthy=True
    )


def get_chronos_status(chronos_client):
    """Gather information about chronos.
    :return: string containing the status
    """
    return run_healthchecks_with_param(
        chronos_client, [assert_chronos_scheduled_jobs, assert_chronos_queued_jobs]
    )


def get_kube_status(kube_client: KubeClient,) -> Sequence[HealthCheckResult]:
    """Gather information about Kubernetes.
    :param kube_client: the KUbernetes client
    :return: string containing the status
    """
    return run_healthchecks_with_param(
        kube_client, [assert_kube_deployments, assert_kube_pods_running]
    )


def critical_events_in_outputs(healthcheck_outputs):
    """Given a list of HealthCheckResults return those which are unhealthy.
    """
    return [
        healthcheck
        for healthcheck in healthcheck_outputs
        if healthcheck.healthy is False
    ]


def generate_summary_for_check(name, ok):
    """Given a check name and a boolean indicating if the service is OK, return
    a formatted message.
    """
    status = PaastaColors.green("OK") if ok is True else PaastaColors.red("CRITICAL")
    summary = f"{name} Status: {status}"
    return summary


def status_for_results(healthcheck_results):
    """Given a list of HealthCheckResult tuples, return the ok status
    for each one.
    :param healthcheck_results: a list of HealthCheckResult tuples
    :returns: a list of booleans.
    """
    return [result.healthy for result in healthcheck_results]


def print_results_for_healthchecks(summary, ok, results, verbose, indent=2):
    paasta_print(summary)
    if verbose >= 1:
        for health_check_result in results:
            if health_check_result.healthy:
                print_with_indent(health_check_result.message, indent)
            else:
                print_with_indent(PaastaColors.red(health_check_result.message), indent)
    elif not ok:
        unhealthy_results = critical_events_in_outputs(results)
        for health_check_result in unhealthy_results:
            print_with_indent(PaastaColors.red(health_check_result.message), indent)


def healthcheck_result_resource_utilization_pair_for_resource_utilization(
    utilization, threshold
):
    """Given a ResourceUtilization, produce a tuple of (HealthCheckResult, ResourceUtilization),
    where that HealthCheckResult describes the 'health' of a given utilization.
    :param utilization: a ResourceUtilization tuple
    :param threshold: a threshold which decides the health of the given ResourceUtilization
    :returns: a tuple of (HealthCheckResult, ResourceUtilization)
    """
    return (
        healthcheck_result_for_resource_utilization(utilization, threshold),
        utilization,
    )


def format_table_column_for_healthcheck_resource_utilization_pair(
    healthcheck_utilization_pair
):
    """Given a tuple of (HealthCheckResult, ResourceUtilization), return a
    string representation of the ResourceUtilization such that it is formatted
    according to the value of HealthCheckResult.healthy.

    :param healthcheck_utilization_pair: a tuple of (HealthCheckResult, ResourceUtilization)
    :returns: a string representing the ResourceUtilization.
    """
    color_func = (
        PaastaColors.green
        if healthcheck_utilization_pair[0].healthy
        else PaastaColors.red
    )
    utilization = (
        healthcheck_utilization_pair[1].total - healthcheck_utilization_pair[1].free
    )
    if int(healthcheck_utilization_pair[1].total) == 0:
        utilization_perc = 100
    else:
        utilization_perc = (
            utilization / float(healthcheck_utilization_pair[1].total) * 100
        )
    if healthcheck_utilization_pair[1].metric not in ["cpus", "gpus"]:
        return color_func(
            "{}/{} ({:.2f}%)".format(
                naturalsize(utilization * 1024 * 1024, gnu=True),
                naturalsize(
                    healthcheck_utilization_pair[1].total * 1024 * 1024, gnu=True
                ),
                utilization_perc,
            )
        )
    else:
        return color_func(
            "{:.2f}/{:.0f} ({:.2f}%)".format(
                utilization, healthcheck_utilization_pair[1].total, utilization_perc
            )
        )


def format_row_for_resource_utilization_healthchecks(healthcheck_utilization_pairs):
    """Given a list of (HealthCheckResult, ResourceUtilization) tuples, return a list with each of those
    tuples represented by a formatted string.

    :param healthcheck_utilization_pairs: a list of (HealthCheckResult, ResourceUtilization) tuples.
    :returns: a list containing a string representation of each (HealthCheckResult, ResourceUtilization) tuple.
    """
    return [
        format_table_column_for_healthcheck_resource_utilization_pair(pair)
        for pair in healthcheck_utilization_pairs
    ]


def get_table_rows_for_resource_info_dict(
    attribute_values, healthcheck_utilization_pairs
):
    """ A wrapper method to join together

    :param attribute: The attribute value and formatted columns to be shown in
    a single row.  :param attribute_value: The value of the attribute
    associated with the row. This becomes index 0 in the array returned.
    :param healthcheck_utilization_pairs: a list of 2-tuples, where each tuple has the elements
    (HealthCheckResult, ResourceUtilization)
    :returns: a list of strings, representing a row in a table to be formatted.
    """
    return attribute_values + format_row_for_resource_utilization_healthchecks(
        healthcheck_utilization_pairs
    )


def reserved_maintenence_resources(resources: MesosResources,):
    return resources.get(MAINTENANCE_ROLE, {"cpus": 0, "mem": 0, "disk": 0, "gpus": 0})



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse

import requests
from service_configuration_lib import read_extra_service_information

from paasta_tools.cli.utils import validate_service_name
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import paasta_print
from paasta_tools.utils import timeout


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "performance-check",
        description="Performs a performance check",
        help="Performs a performance check",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help='Name of service for which you wish to check. Leading "services-", as included in a '
        "Jenkins job name, will be stripped.",
    )
    list_parser.add_argument("-k", "--commit", help=argparse.SUPPRESS)
    list_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="Define a different soa config directory",
    )
    list_parser.set_defaults(command=perform_performance_check)


def load_performance_check_config(service, soa_dir):
    return read_extra_service_information(
        service_name=service, extra_info="performance-check", soa_dir=soa_dir
    )


def submit_performance_check_job(service, soa_dir):
    performance_check_config = load_performance_check_config(service, soa_dir)

    if not performance_check_config:
        paasta_print("No performance-check.yaml. Skipping performance-check.")
        return

    endpoint = performance_check_config.pop("endpoint")
    r = requests.post(url=endpoint, params=performance_check_config)
    r.raise_for_status()
    paasta_print("Posted a submission to the PaaSTA performance-check service.")
    paasta_print(f"Endpoint: {endpoint}")
    paasta_print(f"Parameters: {performance_check_config}")


@timeout()
def perform_performance_check(args):
    service = args.service
    if service.startswith("services-"):
        service = service.split("services-", 1)[1]
    validate_service_name(service, args.soa_dir)

    try:
        submit_performance_check_job(service=service, soa_dir=args.soa_dir)
    except Exception as e:
        paasta_print(
            "Something went wrong with the performance check. Safely bailing. No need to panic."
        )
        paasta_print("Here was the error:")
        paasta_print(str(e))



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

import dulwich.client
import dulwich.errors

from paasta_tools.utils import _run
from paasta_tools.utils import timeout


def _make_determine_wants_func(ref_mutator):
    """Returns a safer version of ref_mutator, suitable for passing as the
    determine_wants argument to dulwich's send_pack method. The returned
    function will not delete or modify any existing refs."""

    def determine_wants(old_refs):
        refs = {k.decode("UTF-8"): v.decode("UTF-8") for k, v in old_refs.items()}
        new_refs = ref_mutator(refs)
        new_refs = {k.encode("UTF-8"): v.encode("UTF-8") for k, v in new_refs.items()}
        new_refs.update(old_refs)  # Make sure we don't delete/modify anything.
        return new_refs

    return determine_wants


def make_force_push_mutate_refs_func(targets, sha):
    """Create a 'force push' function that will inform send_pack that we want
    to mark a certain list of target branches/tags to point to a particular
    git_sha.

    :param targets: List of branches/tags to point at the input sha
    :param sha: The git sha to point the branches/tags at
    :returns: A function to do the ref manipulation that a dulwich client can use"""

    def mutate_refs(refs):
        for target in targets:
            refs[target.encode("UTF-8")] = sha.encode("UTF-8")
        return refs

    return mutate_refs


def create_remote_refs(git_url, ref_mutator, force=False):
    """Creates refs (tags, branches) on a remote git repo.

    :param git_url: the URL or path to the remote git repo.
    :param ref_mutator: A function that determines the new refs to create on
                        the remote repo. This gets passed a dictionary of the
                        remote server's refs in the format {name : hash, ...},
                        and should return a dictionary of the same format.
    :param force: Bool, defaults to false. If true we will overwrite
                  refs even if they are already set.
    :returns: The map of refs, with our changes applied.
    """
    client, path = dulwich.client.get_transport_and_path(git_url)

    if force is False:
        determine_wants = _make_determine_wants_func(ref_mutator)
    else:
        determine_wants = ref_mutator
    # We know we don't need to push any objects.

    def generate_pack_contents(have, want):
        return []

    return client.send_pack(path, determine_wants, generate_pack_contents)


class LSRemoteException(Exception):
    pass


@timeout(
    seconds=60,
    error_message="Timed out connecting to git server, is it reachable from where you are?",
    use_signals=False,
)
def list_remote_refs(git_url):
    """Get the refs from a remote git repo as a dictionary of name->hash."""
    client, path = dulwich.client.get_transport_and_path(git_url)
    try:
        refs = client.fetch_pack(path, lambda refs: [], None, lambda data: None)
        return {k.decode("UTF-8"): v.decode("UTF-8") for k, v in refs.items()}
    except dulwich.errors.HangupException as e:
        raise LSRemoteException(f"Unable to fetch remote refs from {git_url}: {e}")


def get_authors(git_url, from_sha, to_sha):
    """ Gets the list of authors who contributed to a git changeset.
    Currently only supports fetching this in a very "yelpy" way by
    executing a gitolite command """
    matches = re.match("(?P<git_server>.*):(?P<git_repo>.*)", git_url)
    if matches is None:
        return (1, f"could not understand the git url {git_url} for authors detection")
    git_server = matches.group("git_server")
    git_repo = matches.group("git_repo")
    if git_server is None:
        return (
            1,
            f"could not understand the git server in {git_url} for authors detection",
        )
    if git_repo is None:
        return (
            1,
            f"could not understand the git repo in {git_url} for authors detection",
        )

    if "yelpcorp.com" in git_server:
        ssh_command = (
            f"ssh {git_server} authors-of-changeset {git_repo} {from_sha} {to_sha}"
        )
        return _run(command=ssh_command, timeout=5.0)
    else:
        return 1, f"Fetching authors not supported for {git_server}"



import abc
import asyncio
import datetime
import logging
import time
from typing import Callable
from typing import Collection
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import transitions.extensions
from mypy_extensions import TypedDict

log = logging.getLogger(__name__)


class TransitionDefinitionBase(TypedDict):
    """The required fields for TransitionDefinition; see
    https://mypy.readthedocs.io/en/latest/more_types.html#mixing-required-and-non-required-items"""

    trigger: str
    source: Union[str, List[str], Tuple[str, ...]]
    dest: Union[str, List[str], Tuple[str, ...]]


class TransitionDefinition(TransitionDefinitionBase, total=False):
    unless: List[Union[str, Callable]]
    conditions: List[Union[str, Callable]]
    before: Callable


class DeploymentProcess(abc.ABC):
    if TYPE_CHECKING:
        # These attributes need to be defined in this `if TYPE_CHECKING` block, because if they exist at runtime then
        # transitions will refuse to overwrite them.
        state: str

        def trigger(self, *args, **kwargs):
            ...

    run_timeout: Optional[
        float
    ] = None  # in normal operation, this will be None, but this lets tests set a max time.

    def __init__(self,):

        self.event_loop = asyncio.get_event_loop()
        self.finished_event = asyncio.Event(loop=self.event_loop)
        self.timer_running = False

        self.machine = transitions.extensions.LockedMachine(
            model=self,
            states=list(self.states()),
            transitions=list(self.valid_transitions()),
            initial=self.start_state(),
            after_state_change=self.after_state_change,
            before_state_change=self.before_state_change,
            queued=True,
        )

    @abc.abstractmethod
    def status_code_by_state(self) -> Mapping[str, int]:
        raise NotImplementedError()

    @abc.abstractmethod
    def states(self) -> Collection["str"]:
        raise NotImplementedError()

    @abc.abstractmethod
    def valid_transitions(self) -> Iterator[TransitionDefinition]:
        raise NotImplementedError()

    @abc.abstractmethod
    def start_transition(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def start_state(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def notify_users(self, message: str) -> None:
        """Print a log line somewhere that users can see it, e.g. Slack."""
        raise NotImplementedError()

    def is_terminal_state(self, state: str) -> bool:
        return state in self.status_code_by_state()

    def is_finished(self) -> bool:
        return self.finished_event.is_set()

    def finish(self):
        self.finished_event.set()

    def run(self):
        return self.event_loop.run_until_complete(self.run_async())

    async def run_async(self) -> int:
        self.trigger(self.start_transition())
        await asyncio.wait_for(self.finished_event.wait(), timeout=self.run_timeout)
        return self.status_code_by_state().get(self.state, 3)

    def after_state_change(self):
        if self.state in self.status_code_by_state():
            self.event_loop.call_soon_threadsafe(self.finished_event.set)

    def start_timer(self, timeout, trigger, message_verb):
        self.cancel_timer()

        timer_start = time.time()
        timer_end = timer_start + timeout
        formatted_time = datetime.datetime.fromtimestamp(timer_end)

        self.notify_users(
            f"Will {message_verb} in {timeout} seconds, (at {formatted_time})"
        )

        def times_up():
            self.notify_users(f"Time's up, will now {message_verb}.")
            self.trigger(trigger)
            self.timer_handle = None
            self.timer_running = False

        def schedule_callback():
            """Unfortunately, call_at is not threadsafe, and there's no call_at_threadsafe, so we need to schedule the
            call to call_at with call_soon_threadsafe."""
            self.timer_handle = self.event_loop.call_later(timeout, times_up)
            self.timer_trigger = (
                trigger
            )  # This allows cancel_timer to selectively cancel.
            self.timer_timeout = timeout  # saved for restart_timer
            self.timer_message_verb = message_verb  # saved for restart_timer

        self.event_loop.call_soon_threadsafe(schedule_callback)
        self.timer_running = True

    def cancel_timer(self, trigger=None):
        """Cancel the running timer. If trigger is specified, only cancel the timer if its trigger matches."""
        handle = self.get_timer_handle()
        if handle is None:
            self.timer_running = False
            return
        if trigger is None or trigger == self.timer_trigger:
            self.notify_users(f"Countdown to {self.timer_message_verb} cancelled.")
            handle.cancel()
            self.timer_handle = None
            self.timer_running = False
            self.timer_trigger = None
            self.timer_message_verb = None

    def restart_timer(self):
        self.cancel_timer()
        self.start_timer(
            timeout=self.timer_timeout,
            trigger=self.timer_trigger,
            message_verb=self.timer_message_verb,
        )

    def is_timer_running(self) -> bool:
        return self.timer_running

    def get_timer_handle(self):
        try:
            return self.timer_handle
        except AttributeError:
            return None

    def before_state_change(self):
        pass



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import datetime
import json
import logging
from socket import gaierror
from socket import getfqdn
from socket import gethostbyname
from typing import NamedTuple

import a_sync
from dateutil import parser
from pytimeparse import timeparse
from requests import Request
from requests import Session
from requests.exceptions import HTTPError

from paasta_tools.mesos_tools import get_count_running_tasks_on_slave
from paasta_tools.mesos_tools import get_mesos_leader
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.mesos_tools import MESOS_MASTER_PORT
from paasta_tools.utils import to_bytes


log = logging.getLogger(__name__)


class Hostname(NamedTuple):
    host: str
    ip: str


class Credentials(NamedTuple):
    file: str
    principal: str
    secret: str


class Resource(NamedTuple):
    name: str
    amount: int


MAINTENANCE_ROLE = "maintenance"


def base_api():
    """Helper function for making all API requests

    :returns: a function that can be called to make a request
    """
    leader = get_mesos_leader()

    def execute_request(method, endpoint, timeout=(3, 2), **kwargs):
        url = "http://%s:%d%s" % (leader, MESOS_MASTER_PORT, endpoint)
        s = Session()
        s.auth = (get_principal(), get_secret())
        req = Request(method, url, **kwargs)
        prepared = s.prepare_request(req)
        try:
            resp = s.send(prepared, timeout=timeout)
            resp.raise_for_status()
            return resp
        except HTTPError:
            raise HTTPError("Error executing API request calling %s." % url)

    return execute_request


def master_api():
    """Helper function for making API requests to the /master API endpoints

    :returns: a function that can be called to make a request to /master
    """

    def execute_master_api_request(method, endpoint, **kwargs):
        base_api_client = base_api()
        return base_api_client(method, "/master%s" % endpoint, **kwargs)

    return execute_master_api_request


def operator_api():
    def execute_operator_api_request(**kwargs):
        base_api_client = base_api()
        if "headers" in kwargs:
            kwargs["headers"]["Content-Type"] = "application/json"
        else:
            kwargs["headers"] = {"Content-Type": "application/json"}
        data = kwargs.pop("data")
        return base_api_client("POST", "/api/v1", data=json.dumps(data), **kwargs)

    return execute_operator_api_request


def reserve_api():
    """Helper function for making API requests to the /reserve API endpoints

    :returns: a function that can be called to make a request to /reserve
    """

    def execute_reserve_api_request(method, endpoint, **kwargs):
        master_api_client = master_api()
        return master_api_client(method, "/reserve%s" % endpoint, **kwargs)

    return execute_reserve_api_request


def unreserve_api():
    """Helper function for making API requests to the /unreserve API endpoints

    :returns: a function that can be called to make a request to /unreserve
    """

    def execute_unreserve_api_request(method, endpoint, **kwargs):
        master_api_client = master_api()
        return master_api_client(method, "/unreserve%s" % endpoint, **kwargs)

    return execute_unreserve_api_request


def maintenance_api():
    """Helper function for making API requests to the /master/maintenance API endpoints

    :returns: a function that can be called to make a request to /master/maintenance
    """

    def execute_schedule_api_request(method, endpoint, **kwargs):
        master_api_client = master_api()
        return master_api_client(
            method, "/maintenance%s" % endpoint, timeout=(3, 10), **kwargs
        )

    return execute_schedule_api_request


def get_schedule_client():
    """Helper function for making API requests to the /master/maintenance/schedule API endpoints

    :returns: a function that can be called to make a request to /master/maintenance/schedule
    """

    def execute_schedule_api_request(method, endpoint, **kwargs):
        maintenance_api_client = maintenance_api()
        return maintenance_api_client(method, "/schedule%s" % endpoint, **kwargs)

    return execute_schedule_api_request


def get_maintenance_schedule():
    """Makes a GET_MAINTENANCE_SCHEDULE request to the operator api

    :returns: a GET_MAINTENANCE_SCHEDULE response
    """
    client_fn = operator_api()
    return client_fn(data={"type": "GET_MAINTENANCE_SCHEDULE"})


def get_maintenance_status():
    """Makes a GET_MAINTENANCE_STATUS request to the operator api

    :returns: a GET_MAINTENANCE_STATUS response
    """
    client_fn = operator_api()
    return client_fn(data={"type": "GET_MAINTENANCE_STATUS"})


def schedule():
    """Get the Mesos maintenance schedule. This contains hostname/ip mappings and their maintenance window.
    :returns: GET_MAINTENANCE_SCHEDULE response text
    """
    try:
        schedule = get_maintenance_schedule()
    except HTTPError:
        raise HTTPError("Error getting maintenance schedule.")
    return schedule.text


def get_hosts_with_state(state):
    """Helper function to check the maintenance status and return all hosts
    listed as being in a current state

    :param state: State we are interested in ('down_machines' or 'draining_machines')
    :returns: A list of hostnames in the specified state or an empty list if no machines
    """
    try:
        status = get_maintenance_status().json()
        status = status["get_maintenance_status"]["status"]
    except HTTPError:
        raise HTTPError("Error getting maintenance status.")
    if not status or state not in status:
        return []
    if "id" in status[state][0]:
        return [machine["id"]["hostname"] for machine in status[state]]
    else:
        return [machine["hostname"] for machine in status[state]]


def get_draining_hosts():
    """Returns a list of hostnames that are marked as draining

    :returns: a list of strings representing hostnames
    """
    return get_hosts_with_state(state="draining_machines")


def get_down_hosts():
    """Returns a list of hostnames that are marked as down

    :returns: a list of strings representing hostnames
    """
    return get_hosts_with_state(state="down_machines")


def is_host_draining(hostname=getfqdn()):
    """Checks if the specified hostname is marked as draining

    :param hostname: Hostname we want to check if draining (defaults to current host)
    :returns: a boolean representing whether or not the specified hostname is draining
    """
    return hostname in get_draining_hosts()


def is_host_down(hostname=getfqdn()):
    """Checks if the specified hostname is marked as down

    :param hostname: Hostname we want to check if down (defaults to current host)
    :returns: a boolean representing whether or not the specified hostname is down
    """
    return hostname in get_down_hosts()


def get_hosts_forgotten_draining(grace=0):
    """Find hosts that are still marked as draining (rather than down) after the start
    of their maintenance window.
    :param grace: integer number of nanoseconds to allow a host to be left in the draining
    state after the start of its maintenance window before we consider it forgotten.
    :returns: a list of hostnames of hosts forgotten draining
    """
    draining_hosts = get_draining_hosts()
    log.debug("draining_hosts: %s" % draining_hosts)

    hosts_past_maintenance_start = get_hosts_past_maintenance_start(grace=grace)
    log.debug("hosts_past_maintenance_start: %s" % hosts_past_maintenance_start)

    forgotten_draining = list(
        set(draining_hosts).intersection(hosts_past_maintenance_start)
    )
    log.debug("forgotten_draining: %s" % forgotten_draining)

    return forgotten_draining


def are_hosts_forgotten_draining():
    """Quick way to test if there are any forgotten draining hosts.
    :returns: a boolean that is True if there are any forgotten draining
    hosts and False otherwise
    """
    return bool(get_hosts_forgotten_draining())


def get_hosts_forgotten_down(grace=0):
    """Find hosts that are still marked as down (rather than up) after the end
    of their maintenance window.
    :param grace: integer number of nanoseconds to allow a host to be left in the down
    state after the end of its maintenance window before we consider it forgotten.
    :returns: a list of hostnames of hosts forgotten down
    """
    down_hosts = get_down_hosts()
    log.debug("down_hosts: %s" % down_hosts)

    hosts_past_maintenance_end = get_hosts_past_maintenance_end(grace=grace)
    log.debug("hosts_past_maintenance_end: %s" % hosts_past_maintenance_end)

    forgotten_down = list(set(down_hosts).intersection(hosts_past_maintenance_end))
    log.debug("forgotten_down: %s" % forgotten_down)

    return forgotten_down


def are_hosts_forgotten_down():
    """Quick way to test if there are any forgotten down hosts.
    :returns: a boolean that is True if there are any forgotten down
    hosts and False otherwise
    """
    return bool(get_hosts_forgotten_down())


def parse_timedelta(value):
    """Return the delta in nanoseconds.
    :param value: a string containing a time format supported by :mod:`pytimeparse`
    :returns: an integer (or float) representing the specified delta in nanoseconds
    """
    error_msg = "'%s' is not a valid time expression" % value
    try:
        seconds = timeparse.timeparse(value)
    except TypeError:
        raise argparse.ArgumentTypeError(error_msg)
    if not seconds:
        raise argparse.ArgumentTypeError(error_msg)
    return seconds_to_nanoseconds(seconds)


def parse_datetime(value):
    """Return the datetime in nanoseconds.
    :param value: a string containing a datetime supported by :mod:`dateutil.parser`
    :returns: an integer (or float) representing the specified datetime in nanoseconds
    """
    error_msg = "'%s' is not a valid datetime expression" % value
    try:
        dt = parser.parse(value)
    except Exception:
        raise argparse.ArgumentTypeError(error_msg)
    if not dt:
        raise argparse.ArgumentTypeError(error_msg)
    return datetime_to_nanoseconds(dt)


def datetime_seconds_from_now(seconds):
    """Given a number of seconds, returns a datetime object representing that number of seconds in the future from the
    current time.
    :param seconds: an integer representing a certain number of seconds
    :returns: a datetime.timedelta representing now + the specified number of seconds
    """
    return now() + datetime.timedelta(seconds=seconds)


def now():
    """Returns a datetime object representing the current time

    :returns: a datetime.datetime object representing the current time
    """
    return datetime.datetime.now()


def seconds_to_nanoseconds(seconds):
    """Convert the specified number of seconds to nanoseconds
    :param seconds: an integer representing a certain number of seconds
    :returns: an integer (or float) representation of the specified number of seconds as nanoseconds
    """
    return seconds * 1000000000


def datetime_to_nanoseconds(dt):
    """Convert the provided datetime object into nanoseconds

    :returns: an integer (or float) representation of the specified datetime as nanoseconds
    """
    return seconds_to_nanoseconds(int(dt.strftime("%s")))


def build_maintenance_payload(hostnames, maint_type):
    """Creates the JSON payload necessary to bring the specified hostnames up/down for maintenance.
    :param hostnames: a list of hostnames
    :returns: a dictionary representing the list of machines to bring up/down for maintenance
    """
    return {
        "type": maint_type.upper(),
        maint_type.lower(): {"machines": get_machine_ids(hostnames)},
    }


def hostnames_to_components(hostnames, resolve=False):
    """Converts a list of 'host[|ip]' entries into namedtuples containing 'host' and 'ip' attributes,
    optionally performing a DNS lookup to resolve the hostname into an IP address
    :param hostnames: a list of hostnames where each hostname can be of the form 'host[|ip]'
    :param resolve: boolean representing whether to lookup the IP address corresponding to the hostname via DNS
    :returns: a namedtuple containing the hostname and IP components
    """

    components = []
    for hostname in hostnames:
        # This is to allow specifying a hostname as "hostname|ipaddress"
        # to avoid querying DNS for the IP.
        if "|" in hostname:
            (host, ip) = hostname.split("|")
            components.append(Hostname(host=host, ip=ip))
        else:
            try:
                ip = gethostbyname(hostname) if resolve else None
            except gaierror:
                log.error(f"Failed to resolve IP for {hostname}, continuing regardless")
                continue
            components.append(Hostname(host=hostname, ip=ip))
    return components


def get_machine_ids(hostnames):
    """Helper function to convert a list of hostnames into a JSON list of hostname/ip pairs.
    :param hostnames: a list of hostnames
    :returns: a dictionary representing the list of machines to bring up/down for maintenance
    """
    machine_ids = []
    components = hostnames_to_components(hostnames, resolve=True)
    for component in components:
        machine_id = {"hostname": component.host, "ip": component.ip}
        machine_ids.append(machine_id)
    return machine_ids


def build_reservation_payload(resources):
    """Creates the JSON payload needed to dynamically (un)reserve resources in mesos.
    :param resources: list of Resource named tuples specifying the name and amount of the resource to (un)reserve
    :returns: a dictionary that can be sent to Mesos to (un)reserve resources
    """
    payload = []
    for resource in resources:
        payload.append(
            {
                "name": resource.name,
                "type": "SCALAR",
                "scalar": {"value": resource.amount},
                "role": MAINTENANCE_ROLE,
                "reservation": {"principal": get_principal()},
            }
        )
    return payload


def build_maintenance_schedule_payload(
    hostnames, start=None, duration=None, drain=True
):
    """Creates the JSON payload needed to (un)schedule maintenance on the specified hostnames.
    :param hostnames: a list of hostnames
    :param start: the time to start the maintenance, represented as number of nanoseconds since the epoch
    :param duration: length of the maintenance window, represented as number of nanoseconds since the epoch
    :param drain: boolean to note whether we are draining (True) the specified hosts or undraining (False) them
    :returns: a dictionary that can be sent to Mesos to (un)schedule maintenance
    """
    schedule = get_maintenance_schedule().json()["get_maintenance_schedule"]["schedule"]
    machine_ids = get_machine_ids(hostnames)

    if drain:
        unavailability = dict()
        unavailability["start"] = dict()
        unavailability["start"]["nanoseconds"] = int(start)
        unavailability["duration"] = dict()
        unavailability["duration"]["nanoseconds"] = int(duration)

        window = dict()
        window["machine_ids"] = machine_ids
        window["unavailability"] = unavailability

    if schedule:
        for existing_window in schedule["windows"]:
            for existing_machine_id in existing_window["machine_ids"]:
                # If we already have a maintenance window scheduled for one of the hosts,
                # replace it with the new window.
                if existing_machine_id in machine_ids:
                    existing_window["machine_ids"].remove(existing_machine_id)
                    if not existing_window["machine_ids"]:
                        schedule["windows"].remove(existing_window)
        if drain:
            windows = schedule["windows"] + [window]
        else:
            windows = schedule["windows"]
    elif drain:
        windows = [window]
    else:
        windows = []

    payload = dict()
    payload["windows"] = windows

    return {
        "type": "UPDATE_MAINTENANCE_SCHEDULE",
        "update_maintenance_schedule": {"schedule": payload},
    }


def load_credentials(mesos_secrets="/nail/etc/mesos-slave-secret"):
    """Loads the mesos-slave credentials from the specified file. These credentials will be used for all
    maintenance API requests.
    :param mesos_secrets: optional argument specifying the path to the file containing the mesos-slave credentials
    :returns: a tuple of the form (username, password)
    """
    try:
        with open(mesos_secrets) as data_file:
            data = json.load(data_file)
    except EnvironmentError:
        log.error(
            "maintenance calls must be run on a Mesos slave containing valid credentials (%s)"
            % mesos_secrets
        )
        raise
    try:
        username = data["principal"]
        password = data["secret"]
    except KeyError:
        log.error(
            "%s does not contain Mesos slave credentials in the expected format. "
            "See http://mesos.apache.org/documentation/latest/authentication/ for details"
            % mesos_secrets
        )
        raise
    return Credentials(file=mesos_secrets, principal=username, secret=password)


def get_principal(mesos_secrets="/nail/etc/mesos-slave-secret"):
    """Helper function to get the principal from the mesos-slave credentials
    :param mesos_secrets: optional argument specifying the path to the file containing the mesos-slave credentials
    :returns: a string containing the principal/username
    """
    return load_credentials(mesos_secrets).principal


def get_secret(mesos_secrets="/nail/etc/mesos-slave-secret"):
    """Helper function to get the secret from the mesos-slave credentials
    :param mesos_secrets: optional argument specifying the path to the file containing the mesos-slave credentials
    :returns: a string containing the secret/password
    """
    return load_credentials(mesos_secrets).secret


def _make_request_payload(slave_id, reservation_payload):
    return {
        "slaveId": slave_id.encode("UTF-8"),
        # We used to_bytes here since py2 json doesn't have a well defined
        # return type.  When moving to python 3, replace with .encode()
        "resources": to_bytes(json.dumps(reservation_payload)).replace(b"+", b"%20"),
    }


def _make_operator_reservation_request_payload(slave_id, payload, request_type):
    return {
        "type": request_type.upper(),
        request_type.lower(): {"agent_id": {"value": slave_id}},
        "resources": payload,
    }


def reserve(slave_id, resources):
    """Dynamically reserve resources in marathon to prevent tasks from using them.
    :param slave_id: the id of the mesos slave
    :param resources: list of Resource named tuples specifying the name and amount of the resource to (un)reserve
    :returns: boolean where 0 represents success and 1 is a failure
    """
    log.info(f"Dynamically reserving resources on {slave_id}: {resources}")
    payload = _make_operator_reservation_request_payload(
        slave_id=slave_id,
        payload=build_reservation_payload(resources),
        request_type="reserve_resources",
    )
    client_fn = operator_api()
    try:
        print(payload)
        reserve_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error adding dynamic reservation.")
    return reserve_output


def unreserve(slave_id, resources):
    """Dynamically unreserve resources in marathon to allow tasks to using them.
    :param slave_id: the id of the mesos slave
    :param resources: list of Resource named tuples specifying the name and amount of the resource to (un)reserve
    :returns: boolean where 0 represents success and 1 is a failure
    """
    log.info(f"Dynamically unreserving resources on {slave_id}: {resources}")
    payload = _make_operator_reservation_request_payload(
        slave_id=slave_id,
        payload=build_reservation_payload(resources),
        request_type="unreserve_resources",
    )
    client_fn = operator_api()
    try:
        unreserve_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error adding dynamic unreservation.")
    return unreserve_output


def components_to_hosts(components):
    """Convert a list of Component namedtuples to a list of their hosts
    :param components: a list of Component namedtuples
    :returns: list of the hosts associated with each Component
    """
    hosts = []
    for component in components:
        hosts.append(component.host)
    return hosts


def reserve_all_resources(hostnames):
    """Dynamically reserve all available resources on the specified hosts
    :param hostnames: list of hostnames to reserve resources on
    """
    mesos_state = a_sync.block(get_mesos_master().state_summary)
    components = hostnames_to_components(hostnames)
    hosts = components_to_hosts(components)
    known_slaves = [
        slave for slave in mesos_state["slaves"] if slave["hostname"] in hosts
    ]
    for slave in known_slaves:
        hostname = slave["hostname"]
        log.info("Reserving all resources on %s" % hostname)
        slave_id = slave["id"]
        resources = []
        for resource in ["disk", "mem", "cpus", "gpus"]:
            free_resource = (
                slave["resources"][resource] - slave["used_resources"][resource]
            )
            for role in slave["reserved_resources"]:
                free_resource -= slave["reserved_resources"][role][resource]
            resources.append(Resource(name=resource, amount=free_resource))
        try:
            reserve(slave_id=slave_id, resources=resources)
        except HTTPError:
            raise HTTPError(
                f"Failed reserving all of the resources on {hostname} ({slave_id}). Aborting."
            )


def unreserve_all_resources(hostnames):
    """Dynamically unreserve all available resources on the specified hosts
    :param hostnames: list of hostnames to unreserve resources on
    """
    mesos_state = a_sync.block(get_mesos_master().state_summary)
    components = hostnames_to_components(hostnames)
    hosts = components_to_hosts(components)
    known_slaves = [
        slave for slave in mesos_state["slaves"] if slave["hostname"] in hosts
    ]
    for slave in known_slaves:
        hostname = slave["hostname"]
        log.info("Unreserving all resources on %s" % hostname)
        slave_id = slave["id"]
        resources = []
        if MAINTENANCE_ROLE in slave["reserved_resources"]:
            for resource in ["disk", "mem", "cpus", "gpus"]:
                reserved_resource = slave["reserved_resources"][MAINTENANCE_ROLE][
                    resource
                ]
                resources.append(Resource(name=resource, amount=reserved_resource))
            try:
                unreserve(slave_id=slave_id, resources=resources)
            except HTTPError:
                raise HTTPError(
                    f"Failed unreserving all of the resources on {hostname} ({slave_id}). Aborting."
                )


def drain(hostnames, start, duration, reserve_resources=True):
    """Schedules a maintenance window for the specified hosts and marks them as draining.
    :param hostnames: a list of hostnames
    :param start: the time to start the maintenance, represented as number of nanoseconds since the epoch
    :param duration: length of the maintenance window, represented as number of nanoseconds since the epoch
    :param reserve_resources: bool setting to also reserve the free resources on the agent before the drain call
    :returns: None
    """
    log.info("Draining: %s" % hostnames)
    if reserve_resources:
        try:
            reserve_all_resources(hostnames)
        except HTTPError as e:
            log.warning("Failed to reserve resources, will continue to drain: %s" % e)
    payload = build_maintenance_schedule_payload(hostnames, start, duration, drain=True)
    client_fn = operator_api()
    try:
        drain_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error performing maintenance drain.")
    return drain_output


def undrain(hostnames, unreserve_resources=True):
    """Unschedules the maintenance window for the specified hosts and unmarks them as draining. They are ready for
    regular use.
    :param hostnames: a list of hostnames
    :param unreserve_resources: bool setting to also unreserve resources on the agent before the undrain call
    :returns: None
    """
    log.info("Undraining: %s" % hostnames)
    if unreserve_resources:
        try:
            unreserve_all_resources(hostnames)
        except HTTPError as e:
            log.warning(
                "Failed to unreserve resources, will continue to undrain: %s" % e
            )
    payload = build_maintenance_schedule_payload(hostnames, drain=False)
    client_fn = get_schedule_client()
    client_fn = operator_api()
    try:
        undrain_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error performing maintenance undrain.")
    return undrain_output


def down(hostnames):
    """Marks the specified hostnames as being down for maintenance, and makes them unavailable for use.
    :param hostnames: a list of hostnames
    :returns: None
    """
    log.info("Bringing down: %s" % hostnames)
    payload = build_maintenance_payload(hostnames, "start_maintenance")
    client_fn = operator_api()
    try:
        down_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error performing maintenance down.")
    return down_output


def up(hostnames):
    """Marks the specified hostnames as no longer being down for maintenance, and makes them available for use.
    :param hostnames: a list of hostnames
    :returns: None
    """
    log.info("Bringing up: %s" % hostnames)
    payload = build_maintenance_payload(hostnames, "stop_maintenance")
    client_fn = operator_api()
    try:
        up_output = client_fn(data=payload).text
    except HTTPError:
        raise HTTPError("Error performing maintenance up.")
    return up_output


def raw_status():
    """Get the Mesos maintenance status. This contains hostname/ip mappings for hosts that are either marked as being
    down for maintenance or draining.
    :returns: Response Object containing status
    """
    try:
        status = get_maintenance_status()
    except HTTPError:
        raise HTTPError("Error performing maintenance status.")
    return status


def status():
    """Get the Mesos maintenance status. This contains hostname/ip mappings for hosts that are either marked as being
    down for maintenance or draining.
    :returns: Text representation of the status
    """
    return raw_status().text


def friendly_status():
    """Display the Mesos maintenance status in a human-friendly way.
    :returns: Text representation of the human-friendly status
    """
    status = raw_status().json()["get_maintenance_status"]["status"]
    ret = ""
    for machine in status.get("draining_machines", []):
        ret += "{} ({}): Draining\n".format(
            machine["id"]["hostname"], machine["id"]["ip"]
        )
    for machine in status.get("down_machines", []):
        ret += "{} ({}): Down\n".format(machine["hostname"], machine["ip"])
    return ret


def is_host_drained(hostname):
    """Checks if a host has drained successfully by confirming it is
    draining and currently running 0 tasks
    :param hostname: hostname to check
    :returns: True or False
    """
    return (
        is_host_draining(hostname=hostname)
        and get_count_running_tasks_on_slave(hostname) == 0
    )


def is_host_past_maintenance_start(hostname):
    """Checks if a host has reached the start of its maintenance window
    :param hostname: hostname to check
    :returns: True or False
    """
    return hostname in get_hosts_past_maintenance_start()


def is_host_past_maintenance_end(hostname):
    """Checks if a host has reached the end of its maintenance window
    :param hostname: hostname to check
    :returns: True or False
    """
    return hostname in get_hosts_past_maintenance_end()


def get_hosts_past_maintenance_start(grace=0):
    """Get a list of hosts that have reached the start of their maintenance window
    :param grace: integer number of nanoseconds to allow a host to be left in the draining
    state after the start of its maintenance window before we consider it past its maintenance start
    :returns: List of hostnames
    """
    schedules = get_maintenance_schedule().json()["get_maintenance_schedule"][
        "schedule"
    ]
    current_time = datetime_to_nanoseconds(now()) - grace
    ret = []
    if "windows" in schedules:
        for window in schedules["windows"]:
            if window["unavailability"]["start"]["nanoseconds"] < current_time:
                ret += [host["hostname"] for host in window["machine_ids"]]
    log.debug(f"Hosts past maintenance start: {ret}")
    return ret


def get_hosts_past_maintenance_end(grace=0):
    """Get a list of hosts that have reached the end of their maintenance window
    :param grace: integer number of nanoseconds to allow a host to be left in the down
    state after the end of its maintenance window before we consider it past its maintenance end
    :returns: List of hostnames
    """
    schedules = get_maintenance_schedule().json()["get_maintenance_schedule"][
        "schedule"
    ]
    current_time = datetime_to_nanoseconds(now()) - grace
    ret = []
    if "windows" in schedules:
        for window in schedules["windows"]:
            end = (
                window["unavailability"]["start"]["nanoseconds"]
                + window["unavailability"]["duration"]["nanoseconds"]
            )
            if end < current_time:
                ret += [host["hostname"] for host in window["machine_ids"]]
    log.debug(f"Hosts past maintenance end: {ret}")
    return ret



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Usage: ./synapse_srv_namespaces_fact.py

A simple script to enumerate all namespaces as a sorted comma separated
string to stdout, with each entry in the form of full_name:proxy_port.

If a proxy_port isn't defined for a namespace, that namespace is skipped.

Example output: mumble.canary:5019,mumble.main:111,zookeeper.hab:4921

This is nice to use as a facter fact for Synapse stuff!
"""
import sys

from paasta_tools import marathon_tools
from paasta_tools.utils import paasta_print


def main():
    strings = []
    for full_name, config in marathon_tools.get_all_namespaces():
        if "proxy_port" in config:
            strings.append("{}:{}".format(full_name, config["proxy_port"]))
    strings = sorted(strings)
    paasta_print("synapse_srv_namespaces=" + ",".join(strings))
    sys.exit(0)


if __name__ == "__main__":
    main()
